from flask import has_request_context, request, current_app
from frasco.models import db
from frasco.ext import get_extension_state
from frasco.utils import cached_property
from sqlalchemy.ext.declarative import declared_attr
import stripe
import signals
import datetime
import time
import logging


logger = logging.getLogger('frasco.billing.stripe')


class StripeModelMixin(object):
    __stripe_auto_create_customer__ = True
    __stripe_email_property__ = 'email'
    __stripe_customer_expand__ = None

    _stripe_customer_id = db.Column('stripe_customer_id', db.String, index=True)
    has_stripe_source = db.Column(db.Boolean, default=False, index=True)

    @classmethod
    def query_by_stripe_customer(cls, customer_id):
        return cls.query.filter(cls._stripe_customer_id == customer_id)

    @property
    def stripe_customer_id(self):
        return self._stripe_customer_id

    @stripe_customer_id.setter
    def stripe_customer_id(self, value):
        self._stripe_customer_id = value
        self._update_stripe_source()

    @cached_property
    def stripe_customer(self):
        try:
            return stripe.Customer.retrieve(self.stripe_customer_id, expand=self.__stripe_customer_expand__)
        except stripe.error.InvalidRequestError:
            if self.__stripe_auto_create_customer__:
                return self.create_stripe_customer()
            return

    @cached_property
    def stripe_default_source(self):
        if not self.stripe_customer_id:
            return
        default_id = self.stripe_customer.default_source
        if default_id:
            return self.stripe_customer.sources.retrieve(default_id)

    def create_stripe_customer(self, **kwargs):
        kwargs.setdefault('email', getattr(self, self.__stripe_email_property__, None))
        kwargs.setdefault('expand', self.__stripe_customer_expand__)
        customer = stripe.Customer.create(**kwargs)
        self.stripe_customer_id = customer.id if customer else None
        self.__dict__['stripe_customer'] = customer
        self._update_stripe_source()
        logger.info(u'Customer %s for model #%s created' % (customer.id, self.id))
        return customer

    def update_stripe_customer(self, **kwargs):
        customer = self.stripe_customer
        for k, v in kwargs.iteritems():
            setattr(customer, k, v)
        customer.save()
        self.__dict__['stripe_customer'] = customer
        self._update_stripe_source()
        logger.info(u'Customer %s updated' % self.stripe_customer_id)

    def _update_stripe_source(self):
        customer = self.stripe_customer
        self.has_stripe_source = customer.default_source is not None \
            if customer and not getattr(customer, 'deleted', False) else False
        signals.model_source_updated.send(self)

    def delete_stripe_customer(self):
        if self.stripe_customer_id:
            try:
                self.stripe_customer.delete()
                logger.info(u'Customer %s deleted' % self.stripe_customer_id)
            except stripe.error.InvalidRequestError as e:
                if u'No such customer' not in unicode(e):
                    logger.warning(u"Customer %s that was to be deleted didn't exist anymore in Stripe" % self.stripe_customer_id)
                    raise
        self.stripe_customer_id = None
        self.__dict__.pop('stripe_customer', None)

    def add_stripe_source(self, token=None, **source_details):
        self.stripe_customer.sources.create(source=token or source_details)
        self.__dict__.pop('stripe_customer', None)
        self._update_stripe_source()
        logger.info(u'Added new stripe source to %s' % self.stripe_customer_id)

    def remove_stripe_source(self, source_id=None):
        if not source_id:
            source_id = self.stripe_customer.default_source
        try:
            source = self.stripe_customer.sources.retrieve(source_id)
        except stripe.error.InvalidRequestError:
            return
        source.delete()
        self.__dict__.pop('stripe_customer', None)
        self._update_stripe_source()
        logger.info(u'Removed stripe source from %s' % self.stripe_customer_id)

    def charge_stripe_customer(self, amount, **kwargs):
        kwargs['customer'] = self.stripe_customer_id
        kwargs.setdefault('currency', get_extension_state('frasco_stripe').options['default_currency'])
        return stripe.Charge.create(amount=amount, **kwargs)


class StripeSubscriptionModelMixin(StripeModelMixin):
    __stripe_subscription_expand__ = None
    stripe_subscription_id = db.Column(db.String, index=True)
    
    @declared_attr
    def plan_name(cls):
        return db.deferred(db.Column(db.String, index=True), group='stripe_plan')
    
    @declared_attr        
    def plan_status(cls):
        return db.deferred(db.Column(db.String, default='trialing', index=True), group='stripe_plan')
    
    @declared_attr        
    def plan_last_charged_at(cls):
        return db.deferred(db.Column(db.DateTime), group='stripe_plan')
    
    @declared_attr        
    def plan_last_charge_amount(cls):
        return db.deferred(db.Column(db.Float), group='stripe_plan')
    
    @declared_attr        
    def plan_last_charge_successful(cls):
        return db.deferred(db.Column(db.Boolean, default=True, index=True), group='stripe_plan')
    
    @declared_attr        
    def plan_next_charge_at(cls):
        return db.deferred(db.Column(db.DateTime, index=True), group='stripe_plan')

    @classmethod
    def query_by_stripe_subscription(cls, subscription_id):
        return cls.query.filter(cls.stripe_subscription_id == subscription_id)

    @cached_property
    def stripe_subscription(self):
        if not self.stripe_customer_id or not self.stripe_subscription_id:
            return
        try:
            return self.stripe_customer.subscriptions.retrieve(self.stripe_subscription_id, expand=self.__stripe_subscription_expand__)
        except stripe.error.InvalidRequestError:
            return

    def create_stripe_customer(self, trial_end=None, coupon=None, tax_percent=None, **kwargs):
        state = get_extension_state('frasco_stripe')
        if 'plan' in kwargs:
            kwargs.update(dict(trial_end=_format_trial_end(trial_end),
                coupon=coupon, tax_percent=tax_percent))

        customer = super(StripeSubscriptionModelMixin, self).create_stripe_customer(**kwargs)

        if 'plan' in kwargs:
            subscription = customer.subscriptions.data[0]
            self._update_stripe_subscription(subscription)
        elif state.options['default_plan']:
            self.subscribe_stripe_plan(state.options['default_plan'], trial_end=trial_end,
                coupon=coupon, tax_percent=tax_percent)
                
        return customer

    def delete_stripe_customer(self):
        super(StripeSubscriptionModelMixin, self).delete_stripe_customer()
        if self.stripe_subscription_id:
            self._update_stripe_subscription(False)

    def subscribe_stripe_plan(self, plan=None, quantity=1, **kwargs):
        state = get_extension_state('frasco_stripe')
        if not plan:
            plan = state.options['default_plan']
        if self.plan_name == plan:
            return

        params = dict(plan=plan, quantity=quantity,
            trial_end=_format_trial_end(kwargs.pop('trial_end', None)),
            expand=self.__stripe_subscription_expand__)
        params.update(kwargs)
        if 'tax_percent' not in kwargs and state.options['default_subscription_tax_percent']:
            params['tax_percent'] = state.options['default_subscription_tax_percent']

        subscription = self.stripe_customer.subscriptions.create(**params)
        self._update_stripe_subscription(subscription)
        logger.info(u'Subscribed customer %s to plan %s' % (self.stripe_customer_id, plan))
        return subscription

    def update_stripe_subscription(self, **kwargs):
        subscription = self.stripe_subscription
        for k, v in kwargs.iteritems():
            setattr(subscription, k, v)
        subscription.save()
        self._update_stripe_subscription(subscription)
        logger.info(u'Subscription %s updated for %s' % (self.stripe_subscription_id, self.stripe_customer_id))

    def cancel_stripe_subscription(self):
        self.stripe_subscription.delete()
        self._update_stripe_subscription(False)
        logger.info(u'Subscription %s cancelled for %s' % (self.stripe_subscription_id, self.stripe_customer_id))

    def _update_stripe_subscription(self, subscription=None):
        if subscription is None:
            if self.stripe_customer.subscriptions.total_count > 0:
                subscription = self.stripe_customer.subscriptions.data[0]
        prev_plan = self.plan_name
        prev_status = self.plan_status

        if subscription:
            self.stripe_subscription_id = subscription.id
            self.__dict__['stripe_subscription'] = subscription
            self.plan_name = subscription.plan.id
            self.plan_status = subscription.status
            if self.plan_status == 'trialing':
                self.plan_next_charge_at = datetime.datetime.fromtimestamp(subscription.trial_end)
            elif subscription.current_period_end:
                self.plan_next_charge_at = datetime.datetime.fromtimestamp(subscription.current_period_end)
            else:
                self.plan_next_charge_at = None
        else:
            self.stripe_subscription_id = None
            self.__dict__.pop('stripe_subscription', None)
            self.plan_name = None
            self.plan_status = 'canceled'
            self.plan_next_charge_at = None

        signals.model_subscription_updated.send(self, prev_plan=prev_plan, prev_status=prev_status)

    def update_last_stripe_subscription_charge(self, invoice):
        self.plan_last_charged_at = datetime.datetime.fromtimestamp(invoice.date)
        self.plan_last_charge_amount = invoice.total / 100
        self.plan_last_charge_successful = invoice.paid
        if invoice.paid:
            self.plan_next_charge_at = datetime.datetime.fromtimestamp(self.stripe_subscription.current_period_end)
        elif invoice.next_payment_attempt:
            self.plan_next_charge_at = datetime.datetime.fromtimestamp(invoice.next_payment_attempt)
        else:
            self.plan_next_charge_at = None
        signals.model_last_charge_updated.send(self)

    def is_stripe_subscription_valid(self):
        return self.plan_name and (\
            self.plan_status in ('trialing', 'past_due') or \
            (self.plan_status == 'active' and self.has_stripe_source))


class StripeBillingInfoModelMixin(object):
    __stripe_has_billing_fields__ = True
    __stripe_reset_billing_fields__ = True

    @declared_attr
    def billing_name(cls):
        return db.deferred(db.Column(db.String), group='billing')

    @declared_attr
    def billing_address_line1(cls):
        return db.deferred(db.Column(db.String), group='billing')

    @declared_attr
    def billing_address_line2(cls):
        return db.deferred(db.Column(db.String), group='billing')

    @declared_attr
    def billing_address_city(cls):
        return db.deferred(db.Column(db.String), group='billing')

    @declared_attr
    def billing_address_state(cls):
        return db.deferred(db.Column(db.String), group='billing')

    @declared_attr
    def billing_address_zip(cls):
        return db.deferred(db.Column(db.String), group='billing')

    @declared_attr
    def billing_address_country(cls):
        return db.deferred(db.Column(db.String), group='billing')

    @declared_attr
    def billing_country(cls):
        return db.deferred(db.Column(db.String), group='billing')

    @declared_attr
    def billing_ip_address(cls):
        return db.deferred(db.Column(db.String), group='billing')

    @declared_attr
    def billing_brand(cls):
        return db.deferred(db.Column(db.String), group='billing')

    @declared_attr
    def billing_exp_month(cls):
        return db.deferred(db.Column(db.String), group='billing')

    @declared_attr
    def billing_exp_year(cls):
        return db.deferred(db.Column(db.String), group='billing')

    @declared_attr
    def billing_last4(cls):
        return db.deferred(db.Column(db.String), group='billing')
        
    def _update_stripe_source(self):
        super(StripeBillingInfoModelMixin, self)._update_stripe_source()
        if self.has_stripe_source or self.__stripe_reset_billing_fields__:
            billing_fields = ('name', 'address_line1', 'address_line2', 'address_state', 'address_city',
                'address_zip', 'address_country', 'country', 'brand', 'exp_month', 'exp_year', 'last4')
            source = self.stripe_default_source if self.has_stripe_source else None
            for field in billing_fields:
                setattr(self, 'billing_%s' % field, getattr(source, field) if source else None)
            if has_request_context() and self.has_stripe_source:
                self.billing_ip_address = request.remote_addr


class StripeEUVATModelMixin(object):
    __stripe_has_eu_vat__ = True
    __stripe_eu_auto_vat_country__ = True
    __stripe_eu_auto_vat_rate__ = True
    __stripe_eu_vat_use_address_country__ = False

    def _update_stripe_source(self):
        super(StripeEUVATModelMixin, self)._update_stripe_source()
        if self.__stripe_eu_auto_vat_country__:
            country = None
            if self.has_stripe_source:
                if self.__stripe_eu_vat_use_address_country__:
                    country = self.stripe_default_source.address_country
                else:
                    country = self.stripe_default_source.country
            self.eu_vat_country = country
            
        if self.__stripe_eu_auto_vat_rate__ and self.stripe_subscription and self.should_charge_eu_vat():
            self.update_stripe_subscription(tax_percent=self.eu_vat_rate)

    def subscribe_stripe_plan(self, *args, **kwargs):
        if 'tax_percent' not in kwargs and self.__stripe_eu_auto_vat_rate__ and self.should_charge_eu_vat():
            kwargs['tax_percent'] = self.eu_vat_rate
        return super(StripeEUVATModelMixin, self).subscribe_stripe_plan(*args, **kwargs)


def _format_trial_end(trial_end=None):
    state = get_extension_state('frasco_stripe')
    if state.options['debug_trial_period'] and current_app.debug:
        if state.options['debug_trial_period'] == 'now':
            return 'now'
        else:
            trial_end = datetime.datetime.now() + \
                datetime.timedelta(days=state.options['debug_trial_period'])
    if isinstance(trial_end, datetime.datetime):
        if trial_end <= datetime.datetime.now():
            return 'now'
        return int(time.mktime(trial_end.timetuple()))
    return trial_end



from flask import current_app, g
from frasco.ext import get_extension_state, has_extension
from frasco.upload.backend import StorageBackend, RemoteOpenStorageBackendMixin
from frasco.upload.utils import save_uploaded_file_temporarly, generate_filename
from frasco.tasks import enqueue_task
import boto3
import os
import urllib


AWS_CLIENT_OPTIONS = ('aws_access_key_id', 'aws_secret_access_key',
                        'aws_session_token', 'region_name', 'endpoint_url',
                        'config')


class S3StorageBackend(RemoteOpenStorageBackendMixin, StorageBackend):
    default_options = {'filename_prefix': '',
                       'acl': 'public-read',
                       'async': False,
                       'signed_url': False,
                       'signed_url_ttl': 3600,
                       'set_content_disposition_header_with_filename': True,
                       'ignore_images_content_disposition_header': True,
                       'charset': None}

    def save(self, file, filename, force_sync=False):
        kwargs = dict(filename=filename, content_disposition_filename=file.filename,
            bucket=self.options.get('bucket'), acl=self.options['acl'],
            prefix=self.options['filename_prefix'], backend=self.name)

        if not force_sync and self.options.get('async') and has_extension('frasco_tasks'):
            tmpname = save_uploaded_file_temporarly(file, filename)
            enqueue_task(upload_file_to_s3, stream_or_filename=tmpname, mimetype=file.mimetype,
                delete_source=True, **kwargs)
        else:
            upload_file_to_s3(file, **kwargs)

    def url_for(self, filename, **kwargs):
        bucket = self.options.get('bucket')
        if self.options['signed_url']:
            client = get_s3_client(self.name)
            return client.generate_presigned_url('get_object',
                Params={'Bucket': bucket, 'Key': filename},
                ExpiresIn=self.options['signed_url_ttl'])
        return 'https://%s.s3.amazonaws.com/%s' % (bucket, filename)

    def delete(self, filename, force_sync=False):
        if not force_sync and self.options['async'] and has_extension('frasco_tasks'):
            enqueue_task('delete_s3_file', filename=filename, backend=self.name)
        else:
            delete_s3_file(filename, backend=self.name)


def get_s3_options(backend=None):
    return get_extension_state('frasco_upload').get_backend(backend or 's3').options


def get_s3_client(backend=None, use_cached=True):
    if use_cached and "boto_s3_client" in g:
        return g.boto_s3_client
    
    options = get_s3_options(backend)
    kwargs = {}
    if options.get('access_key'):
        kwargs['aws_access_key_id'] = options['access_key']
    if options.get('secret_key'):
        kwargs['aws_secret_access_key'] = options['secret_key']
    for k in AWS_CLIENT_OPTIONS:
        if options.get(k):
            kwargs[k] = options[k]

    client = boto3.client('s3', **kwargs)
    if use_cached:
        g.boto_s3_client = client
    return client


def get_s3_bucket_and_object(filename, backend=None, bucket=None, prefix=None):
    options = get_s3_options(backend)
    if not bucket:
        assert 'bucket' in options, "Missing bucket option"
        bucket = options['bucket']
    prefix = prefix or options['filename_prefix']
    return bucket, prefix + filename


def _generate_object_headers(filename, mimetype=None, charset=None, content_disposition_filename=None, backend=None):
    options = get_s3_options(backend)
    headers = {}

    set_content_disposition = options['set_content_disposition_header_with_filename']
    if options['ignore_images_content_disposition_header'] and mimetype and mimetype.startswith('image/'):
        set_content_disposition = False
    if set_content_disposition:
        if not content_disposition_filename:
            content_disposition_filename = filename
        if isinstance(content_disposition_filename, unicode):
            content_disposition_filename = content_disposition_filename.encode("utf-8")
        headers['Content-Disposition'] = 'attachment;filename="%s"' % urllib.quote(content_disposition_filename)

    if mimetype:
        headers['Content-Type'] = mimetype

    if charset or options['charset']:
        if 'Content-Type' not in headers:
            headers['Content-Type'] = 'binary/octet-stream' # S3 default mimetype
        headers['Content-Type'] += '; charset=%s' % (charset or options['charset'])

    return headers


def upload_file_to_s3(stream_or_filename, filename, bucket=None, prefix=None,
                      acl=None, delete_source=False, backend=None, **object_headers_options):
    options = get_s3_options(backend)
    client = get_s3_client(backend)
    bucket, object_name = get_s3_bucket_and_object(filename, backend, bucket, prefix)
    extra_args = {}

    acl = acl or options['acl']
    if acl:
        extra_args['ACL'] = acl

    extra_args.update({k.replace('-', ''): v for k, v in \
        _generate_object_headers(filename, backend=backend, **object_headers_options).items()})

    is_filename = isinstance(stream_or_filename, (str, unicode))
    if is_filename:
        client.upload_file(stream_or_filename, bucket, object_name, ExtraArgs=extra_args)
    else:
        client.upload_fileobj(stream_or_filename, bucket, object_name, ExtraArgs=extra_args)

    if is_filename and delete_source:
        os.unlink(stream_or_filename)


def generate_s3_presigned_upload_url(max_file_size=None, expires_in=3600, bucket=None, prefix=None, acl=None, backend=None):
    options = get_s3_options(backend)
    client = get_s3_client(backend)
    filename = generate_filename('${filename}')
    bucket, object_name = get_s3_bucket_and_object(filename, backend, bucket, prefix)
    fields = {}
    conditions = []

    acl = acl or options['acl']
    if acl:
        fields['acl'] = acl
        conditions.append({'acl': acl})

    if max_file_size:
        conditions.append(["content-length-range", 1, max_file_size])

    r = client.generate_presigned_post(bucket, object_name, fields, conditions, expires_in)
    return r['url'], r['fields']


def delete_s3_file(filename, bucket=None, prefix=None, backend=None):
    client = get_s3_client(backend)
    bucket, object_name = get_s3_bucket_and_object(filename, backend, bucket, prefix)
    client.delete_object(bucket, object_name)



from frasco.ext import *
from flask_cors import CORS
from .service import *


class FrascoApi(Extension):
    name = 'frasco_api'

    def _init_app(self, app, state):
        state.cors = CORS(app, resources={r"/api/*": {"origins": "*"}})



from flask import request
from frasco.ext import *
from frasco.models import db
from frasco.users import get_current_user, is_user_logged_in
import base64
import datetime
import hashlib
import uuid
import functools
import psycopg2
from .service import ApiAuthentificationRequiredError


class ApiKeyModelMixin(object):
    value = db.Column(db.String, index=True)
    last_accessed_at = db.Column(db.DateTime)
    last_accessed_from = db.Column(db.String)
    expires_at = db.Column(db.DateTime)


class FrascoApiKeyAuthentification(Extension):
    name = 'frasco_api_key_auth'

    def _init_app(self, app, state):
        require_extension('frasco_users', app)
        state.Model = state.import_option('model')

        @app.extensions.frasco_users.manager.request_loader
        def load_user_from_request(request):
            api_key = None
            if 'Authorization' in request.headers:
                authz = request.headers['Authorization']
                if authz.startswith('Basic '):
                    try:
                        api_key = base64.b64decode(authz[6:]).split(':')[0]
                    except:
                        return
                elif authz.startswith('Bearer '):
                    api_key = authz[7:]
            elif 'X-Api-Key' in request.headers:
                api_key = request.headers['X-Api-Key']
            if not api_key:
                return

            try:
                key = state.Model.query.filter_by(value=api_key).first()
            except:
                return
            if key:
                now = datetime.datetime.utcnow()
                if key.expires_at and key.expires_at < now:
                    return None
                key.last_accessed_at = now
                key.last_accessed_from = request.remote_addr
                db.session.commit()
                return key.user


def create_api_key(user=None, expires_at=None):
    state = get_extension_state('frasco_api_key_auth')
    if not expires_at and state.options.get('default_key_duration'):
        expires_at = datetime.datetime.now() + datetime.timedelta(
            seconds=state.options['default_key_duration'])
    key = state.Model()
    key.user = get_current_user(user)
    key.value = hashlib.sha1(str(uuid.uuid4())).hexdigest()
    key.expires_at = expires_at
    return key


def auth_required(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if not is_user_logged_in():
            raise ApiAuthentificationRequiredError()
        return func(*args, **kwargs)
    return wrapper


def authenticated_service(service):
    service.decorators.append(auth_required)
    return service



import mimetypes
from werkzeug import FileStorage, secure_filename as wz_secure_filename
from StringIO import StringIO
from frasco.ext import get_extension_state
from tempfile import NamedTemporaryFile, gettempdir
import os
import uuid


__all__ = ('StringFileStorage', 'generate_filename', 'save_uploaded_file_temporarly', 'file_size', 'format_file_size')


mimetypes.init()


class StringFileStorage(FileStorage):
    def __init__(self, data, filename, mimetype=None, **kwargs):
        super(StringFileStorage, self).__init__(StringIO(data), filename, **kwargs)
        if mimetype is None:
            try:
                mimetype = mimetypes.guess_type(filename)[0]
            except:
                pass
        self._mimetype = mimetype

    @property
    def mimetype(self):
        return self._mimetype


def generate_filename(filename, uuid_prefix=None, uuid_prefix_path_separator=None, keep_filename=None,
                      subfolders=None, backend=None, secure_filename=True):

    state = get_extension_state('frasco_upload', must_exist=False)
    if state:
        if uuid_prefix is None:
            uuid_prefix = state.options["uuid_prefixes"]
        if keep_filename is None:
            keep_filename = state.options["keep_filenames"]
        if subfolders is None:
            subfolders = state.options["subfolders"]
        if uuid_prefix_path_separator is None:
            uuid_prefix_path_separator = state.options["uuid_prefix_path_separator"]

    if uuid_prefix and not keep_filename:
        _, ext = os.path.splitext(filename)
        filename = str(uuid.uuid4()) + ext
    else:
        if secure_filename:
            filename = wz_secure_filename(filename)
        if uuid_prefix:
            filename = str(uuid.uuid4()) + ("/" if uuid_prefix_path_separator else "-") + filename

    if subfolders:
        if uuid_prefix:
            parts = filename.split("-", 4)
            filename = os.path.join(os.path.join(*parts[:4]), filename)
        else:
            filename = os.path.join(os.path.join(*filename[:4]), filename)

    if backend:
        if backend is True:
            backend = state.options['default_backend'] if state else None
        if backend:
            filename = backend + '://' + filename

    return filename


def save_uploaded_file_temporarly(file, filename=None, tmp_dir=None):
    state = get_extension_state('frasco_upload', must_exist=False)
    if not tmp_dir:
        tmp_dir = state.options.get('upload_tmp_dir') if state else None
    if filename:
        tmpfilename = os.path.join(tmp_dir or gettempdir(), wz_secure_filename(filename))
    else:
        _, ext = os.path.splitext(file.filename)
        tmp = NamedTemporaryFile(delete=False, suffix=ext, dir=tmp_dir)
        tmp.close()
        tmpfilename = tmp.name
    file.save(tmpfilename)
    return tmpfilename


def file_size(file):
    file.seek(0, os.SEEK_END)
    size = file.tell()
    file.seek(0)
    return size


def format_file_size(size, suffix='B'):
    for unit in ['','K','M','G','T','P','E','Z']:
        if abs(size) < 1024.0:
            return "%3.1f%s%s" % (size, unit, suffix)
        size /= 1024.0
    return "%.1f%s%s" % (size, 'Y', suffix)



from flask import Blueprint, current_app, request
from .signals import stripe_event_signal
import stripe


webhook_blueprint = Blueprint('stripe_webhook', __name__)


@webhook_blueprint.route('/stripe-webhook', methods=['POST'])
def webhook():
    if current_app.extensions.frasco_stripe.options['webhook_validate_event']:
        event = stripe.Event.retrieve(request.json['id'])
    else:
        event = stripe.Event.construct_from(request.json,
            current_app.extensions.frasco_stripe.options['api_key'])
    stripe_event_signal(event.type.replace(".", "_")).send(stripe, stripe_event=event)
    return 'ok'



from flask import current_app, json
from frasco.utils import unknown_value
import re
import functools
import inspect
import logging


__all__ = ('redis_get_set', 'redis_get_set_as_json', 'build_object_key', 'redis_cached_function', 'redis_cached_function_as_json')


logger = logging.getLogger('frasco.redis')


def redis_get_set(key, callback, ttl=None, coerce=None, serializer=None, redis=None, logging=False):
    if not redis:
        redis = current_app.extensions.frasco_redis.connection
    if redis.exists(key):
        value = redis.get(key)
        if logging or (logging is None and current_app.debug):
            logger.debug(u'CACHE HIT: %s' % key)
        if value == 'None':
            return None
        if coerce:
            return coerce(value)
        return value
    if logging or (logging is None and current_app.debug):
        logger.debug(u'CACHE MISS: %s' % key)
    _value = value = callback()
    if serializer:
        _value = serializer(value)
    if _value is None:
        _value = 'None'
    if ttl:
        redis.setex(key, ttl, _value)
    else:
        redis.set(key, _value)
    return value


def redis_get_set_as_json(key, callback, **kwargs):
    kwargs['serializer'] = json.dumps
    kwargs['coerce'] = json.loads
    return redis_get_set(key, callback, **kwargs)


def build_object_key(obj=None, name=None, key=None, at_values=None, values=None, super_key=None):
    cls = None
    if obj:
        super_key = getattr(obj, '__redis_cache_key__', None)
        if inspect.isclass(obj):
            cls = obj
        else:
            cls = obj.__class__
    elif not key:
        raise ValueError('obj or key is needed for build_object_key()')

    if key and '{__super__}' in key and super_key is not None:
        key = key.replace('{__super__}', super_key)
    elif not key and super_key:
        key = super_key
    elif not key:
        key = '%s:{__name__}' % cls.__name__
    if name is None and cls:
        name = cls.__name__
    if values is None:
        values = {}
    else:
        values = dict(**values)

    for attr in re.findall(r'\{(@?[a-z0-9_]+)[^}]*\}', key, re.I):
        value = unknown_value
        if attr == '__name__' and name is not None:
            value = name
        elif attr.startswith('@') and at_values:
            value = at_values.get(attr[1:], '')
        elif obj:
            value = getattr(obj, attr)
        if value is not unknown_value:
            cache_id = getattr(value, '__redis_cache_id__', None)
            if cache_id:
                value = cache_id()
            values[attr] = value
    return key.format(**values)


def redis_cached_function(key, **opts):
    opts.setdefault('logging', None)
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            if callable(key):
                k = key(*args, **kwargs)
            else:
                k = build_object_key(None, func.__name__, key, values=inspect.getcallargs(func, *args, **kwargs))
            return redis_get_set(k, lambda: func(*args, **kwargs), **opts)
        return wrapper
    return decorator


def redis_cached_function_as_json(key, **opts):
    opts['serializer'] = json.dumps
    opts['coerce'] = json.loads
    return redis_cached_function(key, **opts)



from __future__ import absolute_import
from flask import json
from flask.config import Config as FlaskConfig
from frasco.utils import deep_update_dict
import os
import yaml
import errno
import logging


logger = logging.getLogger('frasco')


class Config(FlaskConfig):
    """Subclass of Flask's Config class to add support to load from YAML file
    """

    def from_json(self, filename, silent=False, deep_update=False):
        filename = os.path.join(self.root_path, filename)

        try:
            with open(filename) as json_file:
                obj = json.loads(json_file.read())
        except IOError as e:
            if silent and e.errno in (errno.ENOENT, errno.EISDIR):
                return False
            e.strerror = 'Unable to load configuration file (%s)' % e.strerror
            raise
        return self.from_mapping(obj, _deep_update=deep_update)

    def from_yaml(self, filename, silent=False, deep_update=False):
        filename = os.path.join(self.root_path, filename)

        try:
            with open(filename) as yaml_file:
                obj = yaml.safe_load(yaml_file.read())
        except IOError as e:
            if silent and e.errno in (errno.ENOENT, errno.EISDIR):
                return False
            e.strerror = 'Unable to load configuration file (%s)' % e.strerror
            raise
        return self.from_mapping(obj, _deep_update=deep_update)

    def from_mapping(self, *mapping, **kwargs):
        mappings = []
        if len(mapping) == 1:
            if hasattr(mapping[0], 'items'):
                mappings.append(mapping[0].items())
            else:
                mappings.append(mapping[0])
        elif len(mapping) > 1:
            raise TypeError(
                'expected at most 1 positional argument, got %d' % len(mapping)
            )
        deep_update = kwargs.pop('_deep_update', False)
        mappings.append(kwargs.items())
        for mapping in mappings:
            if deep_update:
                deep_update_dict(self, dict((k.upper(), v) for (k, v) in mapping))
            else:
                for (key, value) in mapping:
                    self[key.upper()] = value
        return True

    def from_file(self, filename, **kwargs):
        if filename.endswith(".py"):
            return self.from_pyfile(filename, **kwargs)
        if filename.endswith(".js") or filename.endswith(".json"):
            return self.from_json(filename, **kwargs)
        if filename.endswith(".yml") or filename.endswith(".yaml"):
            return self.from_yaml(filename, **kwargs)
        raise RuntimeError("Unknown config file extension")


def load_config(app, config_filename='config.yml', env=None, deep_update=False):
    if os.path.exists(config_filename):
        logger.info('Loading config from %s' % config_filename)
        app.config.from_file(config_filename, deep_update=deep_update)
    if env is False:
        return
    env = env or app.config['ENV']
    filename, ext = os.path.splitext(config_filename)
    env_filename = filename + "-" + env + ext
    if os.path.exists(env_filename):
        logger.info('Loading config from %s' % env_filename)
        app.config.from_file(env_filename, deep_update=True)


def update_config_with_env_vars(app, prefix):
    config = {}
    prefix = prefix.upper()
    for k, v in os.environ.iteritems():
        if k.startswith(prefix + "_"):
            config[k[len(prefix)+1:]] = v
    if config:
        logger.info('Using config from environment variables')
        app.config.update(config)



from frasco import request_param
from frasco.api import ApiService, ApiInputError, ApiNotFoundError
import data


eu_vat_service = ApiService('eu_vat', url_prefix='/eu-vat')


@eu_vat_service.route('/rates/<country_code>')
@request_param('country_code', type=str)
def get_vat_rate(country_code):
    try:
        return data.get_vat_rate(country_code)
    except data.EUVATError as e:
        raise ApiInputError(unicode(e))


@eu_vat_service.route('/validate-vat-number', methods=['POST'])
@request_param('vat_number', type=str)
def validate_vat_number(vat_number):
    try:
        return data.validate_vat_number(vat_number, invalid_format_raise_error=True)
    except data.EUVATError as e:
        raise ApiInputError(unicode(e))


@eu_vat_service.route('/exchange-rates/<country_code>', methods=['POST'])
@eu_vat_service.route('/exchange-rates/<country_code>/<src_currency>', methods=['POST'])
@request_param('country_code', type=str)
@request_param('src_currency', type=str)
def get_exchange_rate(country_code, src_currency='EUR'):
    try:
        return data.get_exchange_rate(country_code, src_currency)
    except data.EUVATError as e:
        raise ApiInputError(unicode(e))


@eu_vat_service.route('/check', methods=['POST'])
@request_param('country_code', type=str)
@request_param('vat_number', type=str)
@request_param('amount', type=float)
@request_param('src_currency', type=str)
def check(country_code, vat_number=None, amount=None, src_currency='EUR'):
    if not data.is_eu_country(country_code):
        raise ApiNotFoundError('Not an EU country')
    
    is_vat_number_valid = data.validate_vat_number(vat_number) if vat_number else False
    o = {
        "country": country_code,
        "currency": data.EU_COUNTRIES[country_code],
        "vat_rate": data.get_vat_rate(country_code),
        "vat_number": vat_number,
        "is_vat_number_valid": is_vat_number_valid,
        "should_charge_vat": data.should_charge_vat(country_code, vat_number and is_vat_number_valid),
        "exchange_rate": eu_vat_service.get_exchange_rate(country_code, src_currency),
        "src_currency": src_currency
    }

    if amount:
        rate = 0
        if o['should_charge_vat']:
            rate = o['vat_rate'] / 100
        o.update({"amount": amount,
                    "vat_amount": round(amount * rate, 2),
                    "amount_with_vat": amount + amount * rate,
                    "exchanged_amount_with_vat": round((amount + amount * rate) * o["exchange_rate"], 2)})
                    
    return o
